@@ -1,5 +1,6 @@
 from __future__ import unicode_literals
 
+import json
 from urllib import urlencode
 from urlparse import parse_qs
 
@@ -16,15 +17,10 @@ class BaseOAuthClient(object):
     def __init__(self, provider):
         self.provider = provider
 
-    def get_access_token(self, request):
+    def get_access_token(self, request, callback=None):
         "Fetch access token from callback request."
         raise NotImplementedError('Defined in a sub-class') # pragma: no cover
 
-    def get_callback_url(self, request):
-        "Construct the callback url for a given provider."
-        callback = reverse('allaccess-callback', kwargs={'provider': self.provider.name})
-        return force_unicode(request.build_absolute_uri(callback))
-
     def get_profile_info(self, raw_token):
         "Fetch user profile information."
         try:
@@ -35,13 +31,15 @@ class BaseOAuthClient(object):
         else:
             return response.json if response.json is not None else response.text
 
-    def get_redirect_args(self, request):
+    def get_redirect_args(self, request, callback):
         "Get request parameters for redirect url."
         raise NotImplementedError('Defined in a sub-class') # pragma: no cover
 
-    def get_redirect_url(self, request):
+    def get_redirect_url(self, request, callback, parameters=None):
         "Build authentication redirect url."
-        args = self.get_redirect_args(request)
+        args = self.get_redirect_args(request, callback=callback)
+        additional = parameters or {}
+        args.update(additional)
         params = urlencode(args)
         return '{0}?{1}'.format(self.provider.authorization_url, params)
 
@@ -56,12 +54,14 @@ class BaseOAuthClient(object):
 
 class OAuthClient(BaseOAuthClient):
 
-    def get_access_token(self, request):
+    def get_access_token(self, request, callback=None):
         "Fetch access token from callback request."
-        raw_token = request.session.get(self.session_key, '')
-        if raw_token:
-            data = {'oauth_verifier': request.GET.get('oauth_verifier', '')}
-            callback = self.get_callback_url(request)
+        raw_token = request.session.get(self.session_key, None)
+        verifier = request.GET.get('oauth_verifier', None)
+        if raw_token is not None and verifier is not None:
+            data = {'oauth_verifier': verifier}
+            callback = request.build_absolute_uri(callback or request.path)
+            callback = force_unicode(callback)
             try:
                 response = self.request('post', self.provider.access_token_url, 
                                         token=raw_token, data=data, oauth_callback=callback)
@@ -72,9 +72,9 @@ class OAuthClient(BaseOAuthClient):
                 return response.text
         return None
 
-    def get_request_token(self, request):
+    def get_request_token(self, request, callback):
         "Fetch the OAuth request token. Only required for OAuth 1.0."
-        callback = self.get_callback_url(request)
+        callback = force_unicode(request.build_absolute_uri(callback))
         try:
             response = self.request('post', self.provider.request_token_url, oauth_callback=callback)
         except RequestException:
@@ -83,13 +83,13 @@ class OAuthClient(BaseOAuthClient):
         else:
             return response.text
 
-    def get_redirect_args(self, request):
+    def get_redirect_args(self, request, callback):
         "Get request parameters for redirect url."
-        raw_token = self.get_request_token(request)
+        callback = force_unicode(request.build_absolute_uri(callback))
+        raw_token = self.get_request_token(request, callback)
         token, secret = self.parse_raw_token(raw_token)
         if token is not None and secret is not None:
             request.session[self.session_key] = raw_token
-        callback = self.get_callback_url(request)
         return {
             'oauth_token': token,
             'oauth_callback': callback,
@@ -126,26 +126,30 @@ class OAuthClient(BaseOAuthClient):
 
 class OAuth2Client(BaseOAuthClient):
 
-    def get_access_token(self, request):
+    def get_access_token(self, request, callback=None):
         "Fetch access token from callback request."
-        callback = self.get_callback_url(request)
-        args = {
-            'client_id': self.provider.key,
-            'redirect_uri': callback,
-            'client_secret': self.provider.secret,
-            'code': request.GET.get('code', '')
-        }
+        callback = request.build_absolute_uri(callback or request.path)
+        if 'code' in request.GET:
+            args = {
+                'client_id': self.provider.key,
+                'redirect_uri': callback,
+                'client_secret': self.provider.secret,
+                'code': request.GET['code'],
+                'grant_type': 'authorization_code',
+            }
+        else:
+            return None
         try:
-            response = self.request('get', self.provider.access_token_url, params=args)
+            response = self.request('post', self.provider.access_token_url, params=args)
         except RequestException:
             # TODO: Logging
             return None
         else:
             return response.text
 
-    def get_redirect_args(self, request):
+    def get_redirect_args(self, request, callback):
         "Get request parameters for redirect url."
-        callback = self.get_callback_url(request)
+        callback = request.build_absolute_uri(callback)
         return {
             'client_id': self.provider.key,
             'redirect_uri': callback,
@@ -154,8 +158,14 @@ class OAuth2Client(BaseOAuthClient):
 
     def parse_raw_token(self, raw_token):
         "Parse token and secret from raw token response."
-        qs = parse_qs(raw_token)
-        token = qs.get('access_token', [None])[0]
+        # Load as json first then parse as query string
+        try:
+            token_data = json.loads(raw_token)
+        except ValueError:
+            qs = parse_qs(raw_token)
+            token = qs.get('access_token', [None])[0]
+        else:
+            token = token_data.get('access_token', None)
         return (token, None)
 
     def request(self, method, url, **kwargs):
